from django import template
from django.utils.safestring import mark_safe

from machina.core.db.models import get_model
from machina.core.loading import get_class


Forum = get_model('forum', 'Forum')

PermissionHandler = get_class('forum_permission.handler', 'PermissionHandler')
TrackingHandler = get_class('forum_tracking.handler', 'TrackingHandler')

register = template.Library()


class RecurseTreeForumVisibilityContentNode(template.Node):
    def __init__(self, template_nodes, forums_contents_var):
        self.template_nodes = template_nodes
        self.forums_contents_var = forums_contents_var

    def _render_node(self, context, node):
        bits = []
        context.push()
        for child in node.children:
            bits.append(self._render_node(context, child))
        context['node'] = node
        context['children'] = mark_safe(''.join(bits))
        rendered = self.template_nodes.render(context)
        context.pop()
        return rendered

    def render(self, context):
        forums_contents = self.forums_contents_var.resolve(context)
        roots = forums_contents.top_nodes
        bits = [self._render_node(context, node) for node in roots]
        return ''.join(bits)


@register.tag
def recurseforumcontents(parser, token):
    """ Iterates over the content nodes and renders the contained forum block for each node. """
    bits = token.contents.split()
    forums_contents_var = template.Variable(bits[1])

    template_nodes = parser.parse(('endrecurseforumcontents',))
    parser.delete_first_token()

    return RecurseTreeForumVisibilityContentNode(template_nodes, forums_contents_var)


@register.inclusion_tag('forum/forum_list.html', takes_context=True)
def forum_list(context, forum_visibility_contents):
    """ Renders the considered forum list.

    This will render the given list of forums by respecting the order and the depth of each
    forum in the forums tree.

    Usage::

        {% forum_list my_forums %}

    """
    request = context.get('request')
    tracking_handler = TrackingHandler(request=request)

    data_dict = {
        'forum_contents': forum_visibility_contents,
        'unread_forums': tracking_handler.get_unread_forums_from_list(
            request.user, forum_visibility_contents.forums),
        'user': request.user,
        'request': request,
    }

    root_level = forum_visibility_contents.root_level
    if root_level is not None:
        data_dict['root_level'] = root_level
        data_dict['root_level_middle'] = root_level + 1
        data_dict['root_level_sub'] = root_level + 2

    return data_dict



"""
    Forum permission signal receivers
    =================================

    This module defines signal receivers.

"""

from django.db.models.signals import post_migrate
from django.dispatch import receiver

from machina.core.db.models import get_model
from machina.core.loading import get_class


ForumPermission = get_model('forum_permission', 'ForumPermission')
PermissionConfig = get_class('forum_permission.defaults', 'PermissionConfig')


def create_permissions():
    """ Creates all the permissions from the permission configuration. """
    for config in PermissionConfig.permissions:
        ForumPermission.objects.get_or_create(codename=config['codename'])


@receiver(post_migrate)
def create_global_permissions(sender, **kwargs):
    """ Creates all the permissions from the permission configuration during migrations. """
    if sender.name.endswith('forum_permission'):
        create_permissions()



"""
    Forum model admin definitions
    =============================

    This module defines admin classes used to populate the Django administration dashboard.

"""

from collections import OrderedDict

from django.contrib import admin
from django.contrib.admin import helpers
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.forms.forms import NON_FIELD_ERRORS
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import re_path, reverse
from django.utils.translation import gettext_lazy as _
from mptt.exceptions import InvalidMove

from machina.core.db.models import get_model
from machina.core.loading import get_class
from machina.models.fields import MarkupTextField, MarkupTextFieldWidget


Forum = get_model('forum', 'Forum')
ForumPermission = get_model('forum_permission', 'ForumPermission')
GroupForumPermission = get_model('forum_permission', 'GroupForumPermission')
UserForumPermission = get_model('forum_permission', 'UserForumPermission')

PermissionsForm = get_class('forum.forms', 'PermissionsForm')
PickForumForm = get_class('forum.forms', 'PickForumForm')
PickGroupForm = get_class('forum.forms', 'PickGroupForm')
PickUserForm = get_class('forum.forms', 'PickUserForm')

PermissionConfig = get_class('forum_permission.defaults', 'PermissionConfig')


class ForumAdmin(admin.ModelAdmin):
    """ The Forum model admin. """

    formfield_overrides = {
        MarkupTextField: {'widget': MarkupTextFieldWidget},
    }

    fieldsets = (
        [None, {'fields': ('type', 'parent', 'name', 'description', 'image')}],
        [_('Forum settings'), {'fields': ('display_sub_forum_list', ), 'classes': ('collapse', )}],
        [
            _('Link forum settings'),
            {'fields': ('link', 'link_redirects', ), 'classes': ('collapse', )},
        ],
    )
    list_display = ('name', 'type', 'direct_topics_count', 'direct_posts_count', )
    search_fields = ('name', )

    editpermissions_index_view_template_name = 'admin/forum/forum/editpermissions_index.html'
    editpermissions_user_view_template_name = 'admin/forum/forum/editpermissions_user.html'
    editpermissions_anonymous_user_view_template_name = (
        'admin/forum/forum/editpermissions_anonymous_user.html'
    )
    editpermissions_authenticated_user_view_template_name = (
        'admin/forum/forum/editpermissions_authenticated_user.html'
    )
    editpermissions_group_view_template_name = 'admin/forum/forum/editpermissions_group.html'

    def get_urls(self):
        """ Returns the URLs associated with the admin abstraction. """
        urls = super().get_urls()
        forum_admin_urls = [
            re_path(
                r'^(?P<forum_id>[0-9]+)/move-forum/(?P<direction>up|down)/$',
                self.admin_site.admin_view(self.moveforum_view),
                name='forum_forum_move',
            ),
            re_path(
                r'^edit-global-permissions/$',
                self.admin_site.admin_view(self.editpermissions_index_view),
                name='forum_forum_editpermission_index',
            ),
            re_path(
                r'^edit-global-permissions/user/(?P<user_id>[0-9]+)/$',
                self.admin_site.admin_view(self.editpermissions_user_view),
                name='forum_forum_editpermission_user',
            ),
            re_path(
                r'^edit-global-permissions/user/anonymous/$',
                self.admin_site.admin_view(self.editpermissions_anonymous_user_view),
                name='forum_forum_editpermission_anonymous_user',
            ),
            re_path(
                r'^edit-global-permissions/user/authenticated/$',
                self.admin_site.admin_view(self.editpermissions_authenticated_user_view),
                name='forum_forum_editpermission_authenticated_user',
            ),
            re_path(
                r'^edit-global-permissions/group/(?P<group_id>[0-9]+)/$',
                self.admin_site.admin_view(self.editpermissions_group_view),
                name='forum_forum_editpermission_group',
            ),
            re_path(
                r'^(?P<forum_id>[0-9]+)/edit-permissions/$',
                self.admin_site.admin_view(self.editpermissions_index_view),
                name='forum_forum_editpermission_index',
            ),
            re_path(
                r'^(?P<forum_id>[0-9]+)/edit-permissions/user/(?P<user_id>[0-9]+)/$',
                self.admin_site.admin_view(self.editpermissions_user_view),
                name='forum_forum_editpermission_user',
            ),
            re_path(
                r'^(?P<forum_id>[0-9]+)/edit-permissions/user/anonymous/$',
                self.admin_site.admin_view(self.editpermissions_anonymous_user_view),
                name='forum_forum_editpermission_anonymous_user',
            ),
            re_path(
                r'^(?P<forum_id>[0-9]+)/edit-permissions/user/authenticated/$',
                self.admin_site.admin_view(self.editpermissions_authenticated_user_view),
                name='forum_forum_editpermission_authenticated_user',
            ),
            re_path(
                r'^(?P<forum_id>[0-9]+)/edit-permissions/group/(?P<group_id>[0-9]+)/$',
                self.admin_site.admin_view(self.editpermissions_group_view),
                name='forum_forum_editpermission_group',
            ),
        ]
        return forum_admin_urls + urls

    def get_forum_perms_base_context(self, request, obj=None):
        """ Returns the context to provide to the template for permissions contents. """
        context = {
            'adminform': {'model_admin': self},
            'media': self.media,
            'object': obj,
            'app_label': self.model._meta.app_label,
            'opts': self.model._meta,
            'has_change_permission': self.has_change_permission(request, obj),
        }
        try:
            context.update(self.admin_site.each_context(request))
        except TypeError:  # pragma: no cover
            # Django 1.7 compatibility
            context.update(self.admin_site.each_context())
        except AttributeError:  # pragma: no cover
            pass
        return context

    def moveforum_view(self, request, forum_id, direction):
        """ Moves the given forum toward the requested direction. """
        forum = get_object_or_404(Forum, pk=forum_id)

        # Fetch the target
        target, position = None, None
        if direction == 'up':
            target, position = forum.get_previous_sibling(), 'left'
        elif direction == 'down':
            target, position = forum.get_next_sibling(), 'right'

        # Do the move
        try:
            assert target is not None
            forum.move_to(target, position)
        except (InvalidMove, AssertionError):
            pass
        self.message_user(request, _("'{}' forum successfully moved").format(forum.name))
        return HttpResponseRedirect(reverse('admin:forum_forum_changelist'))

    def editpermissions_index_view(self, request, forum_id=None):
        """ Allows to select how to edit forum permissions.

        The view displays a form to select a user or a group in order to edit its permissions for
        the considered forum.

        """
        forum = get_object_or_404(Forum, pk=forum_id) if forum_id \
            else None

        # Set up the context
        context = self.get_forum_perms_base_context(request, forum)
        context['forum'] = forum
        context['title'] = _('Forum permissions') if forum else _('Global forum permissions')

        can_change_user_perms = (
            request.user.has_perm('forum_permission.add_userforumpermission') or
            request.user.has_perm('forum_permission.change_userforumpermission')
        )
        can_change_group_perms = (
            request.user.has_perm('forum_permission.add_groupforumpermission') or
            request.user.has_perm('forum_permission.change_groupforumpermission')
        )

        # Handles "copy permission from" form
        permissions_copied = False
        if can_change_user_perms and can_change_group_perms:
            if forum and request.method == 'POST':
                forum_form = PickForumForm(request.POST)
                if forum_form.is_valid() and forum_form.cleaned_data['forum']:
                    self._copy_forum_permissions(forum_form.cleaned_data['forum'], forum)
                    self.message_user(request, _('Permissions successfully copied'))
                    permissions_copied = True
                context['forum_form'] = forum_form
            elif forum:
                context['forum_form'] = PickForumForm()

        user_form, user = None, None
        group_form, group = None, None

        # Handles user or group selection
        if request.method == 'POST' and not permissions_copied:
            if can_change_user_perms:
                user_form = PickUserForm(request.POST, admin_site=self.admin_site)
            if can_change_group_perms:
                group_form = PickGroupForm(request.POST, admin_site=self.admin_site)

            # Check if the form was drawn, was valid and was submitted (presence of submitbutton)
            if user_form and user_form.is_valid() and '_select_user' in request.POST:
                user = user_form.cleaned_data.get('user', None) if user_form.cleaned_data else None
                anonymous_user = (
                    user_form.cleaned_data.get('anonymous_user', None)
                    if user_form.cleaned_data else None
                )
                authenticated_user = (
                    user_form.cleaned_data.get('authenticated_user', None)
                    if user_form.cleaned_data else None
                )

                if not user and not anonymous_user and not authenticated_user:
                    user_form._errors[NON_FIELD_ERRORS] = user_form.error_class([
                        _(
                            "Choose either a user ID, the anonymous user or the authenticated user"
                        ),
                    ])
                elif user:
                    # Redirect to user
                    url_kwargs = (
                        {'forum_id': forum.id, 'user_id': user.id}
                        if forum else {'user_id': user.id}
                    )
                    return redirect(
                        reverse('admin:forum_forum_editpermission_user', kwargs=url_kwargs),
                    )
                elif anonymous_user:
                    # Redirect to anonymous user
                    url_kwargs = {'forum_id': forum.id} if forum else {}
                    return redirect(
                        reverse(
                            'admin:forum_forum_editpermission_anonymous_user',
                            kwargs=url_kwargs,
                        ),
                    )
                elif authenticated_user:
                    # Redirect to authenticated user
                    url_kwargs = {'forum_id': forum.id} if forum else {}
                    return redirect(
                        reverse(
                            'admin:forum_forum_editpermission_authenticated_user',
                            kwargs=url_kwargs,
                        ),
                    )

                context['user_errors'] = helpers.AdminErrorList(user_form, [])

            # Check if the form was drawn, was valid and was submitted (presence of submitbutton)
            if group_form and group_form.is_valid() and '_select_group' in request.POST:
                group = group_form.cleaned_data.get('group', None) \
                    if group_form.cleaned_data else None

                if not group:
                    group_form._errors[NON_FIELD_ERRORS] = group_form.error_class(
                        [_('Choose a group ID'), ])
                else:
                    # Redirect to group
                    url_kwargs = (
                        {'forum_id': forum.id, 'group_id': group.id}
                        if forum else {'group_id': group.id}
                    )
                    return redirect(
                        reverse('admin:forum_forum_editpermission_group', kwargs=url_kwargs),
                    )

                context['group_errors'] = helpers.AdminErrorList(group_form, [])
        else:
            if can_change_user_perms:
                user_form = PickUserForm(admin_site=self.admin_site)
            if can_change_group_perms:
                group_form = PickGroupForm(admin_site=self.admin_site)

        context['user_form'] = user_form
        context['group_form'] = group_form

        return render(request, self.editpermissions_index_view_template_name, context)

    def editpermissions_user_view(self, request, user_id, forum_id=None):
        """ Allows to edit user permissions for the considered forum.

        The view displays a form to define which permissions are granted for the given user for the
        considered forum.

        """
        user_model = get_user_model()
        user = get_object_or_404(user_model, pk=user_id)
        forum = get_object_or_404(Forum, pk=forum_id) if forum_id else None

        # Set up the context
        context = self.get_forum_perms_base_context(request, forum)
        context['forum'] = forum
        context['title'] = '{} - {}'.format(_('Forum permissions'), user)
        context['form'] = self._get_permissions_form(
            request, UserForumPermission, {'forum': forum, 'user': user},
        )

        return render(request, self.editpermissions_user_view_template_name, context)

    def editpermissions_anonymous_user_view(self, request, forum_id=None):
        """ Allows to edit anonymous user permissions for the considered forum.

        The view displays a form to define which permissions are granted for the anonymous user for
        the considered forum.

        """
        forum = get_object_or_404(Forum, pk=forum_id) if forum_id else None

        # Set up the context
        context = self.get_forum_perms_base_context(request, forum)
        context['forum'] = forum
        context['title'] = '{} - {}'.format(_('Forum permissions'), _('Anonymous user'))
        context['form'] = self._get_permissions_form(
            request, UserForumPermission, {'forum': forum, 'anonymous_user': True},
        )

        return render(request, self.editpermissions_anonymous_user_view_template_name, context)

    def editpermissions_authenticated_user_view(self, request, forum_id=None):
        """ Allows to edit authenticated user permissions for the considered forum.

        The view displays a form to define which permissions are granted for the authenticated,
        non-specific, user for the considered forum.

        """
        forum = get_object_or_404(Forum, pk=forum_id) if forum_id else None

        # Set up the context
        context = self.get_forum_perms_base_context(request, forum)
        context['forum'] = forum
        context['title'] = '{} - {}'.format(_('Forum permissions'), _('Authenticated user'))
        context['form'] = self._get_permissions_form(
            request, UserForumPermission, {'forum': forum, 'authenticated_user': True},
        )

        return render(request, self.editpermissions_authenticated_user_view_template_name, context)

    def editpermissions_group_view(self, request, group_id, forum_id=None):
        """ Allows to edit group permissions for the considered forum.

        The view displays a form to define which permissions are granted for the given group for the
        considered forum.

        """
        group = get_object_or_404(Group, pk=group_id)
        forum = get_object_or_404(Forum, pk=forum_id) if forum_id else None

        # Set up the context
        context = self.get_forum_perms_base_context(request, forum)
        context['forum'] = forum
        context['title'] = '{} - {}'.format(_('Forum permissions'), group)
        context['form'] = self._get_permissions_form(
            request, GroupForumPermission, {'forum': forum, 'group': group},
        )

        return render(request, self.editpermissions_group_view_template_name, context)

    def _get_permissions_form(self, request, permission_model, filter_kwargs):
        # Fetch the permissions
        editable_permissions = sorted(
            ForumPermission.objects.all(), key=lambda p: p.name,
        )
        granted_permissions = (
            permission_model.objects.filter(
                permission__in=editable_permissions, has_perm=True, **filter_kwargs
            )
            .values_list('permission__codename', flat=True)
        )
        non_granted_permissions = (
            permission_model.objects.filter(
                permission__in=editable_permissions, has_perm=False, **filter_kwargs
            )
            .values_list('permission__codename', flat=True)
        )

        permissions_dict = OrderedDict()
        for p in editable_permissions:
            if p.codename in granted_permissions:
                perm_state = PermissionsForm.PERM_GRANTED
            elif p.codename in non_granted_permissions:
                perm_state = PermissionsForm.PERM_NOT_GRANTED
            else:
                perm_state = PermissionsForm.PERM_NOT_SET
            permissions_dict[p.codename] = (p, perm_state)

        if request.method == 'POST':
            form = PermissionsForm(request.POST, permissions_dict=permissions_dict)
            if form.is_valid():
                for codename, value in form.cleaned_data.items():
                    try:
                        perm = permission_model.objects.get(
                            permission=permissions_dict[codename][0], **filter_kwargs
                        )
                    except permission_model.DoesNotExist:
                        if value == PermissionsForm.PERM_NOT_SET:
                            continue
                        perm = permission_model.objects.create(
                            permission=permissions_dict[codename][0], **filter_kwargs
                        )

                    if value == PermissionsForm.PERM_NOT_SET:
                        perm.delete()
                        continue

                    perm.has_perm = (value == PermissionsForm.PERM_GRANTED)
                    perm.save()

                self.message_user(request, _('Permissions successfully applied'))
        else:
            form = PermissionsForm(permissions_dict=permissions_dict)

        return form

    def _copy_forum_permissions(self, forum_from, forum_to):
        user_perms = UserForumPermission.objects.filter(forum=forum_from)
        group_perms = GroupForumPermission.objects.filter(forum=forum_from)

        for perm in user_perms:
            try:
                new_perm = UserForumPermission.objects.get(
                    permission=perm.permission, forum=forum_to,
                    user=perm.user, anonymous_user=perm.anonymous_user,
                )
            except UserForumPermission.DoesNotExist:
                new_perm = UserForumPermission(
                    permission=perm.permission, forum=forum_to,
                    user=perm.user, anonymous_user=perm.anonymous_user,
                    authenticated_user=perm.authenticated_user
                )
            new_perm.has_perm = perm.has_perm
            new_perm.save()

        for perm in group_perms:
            try:
                new_perm = GroupForumPermission.objects.get(
                    permission=perm.permission, forum=forum_to,
                    group=perm.group,
                )
            except GroupForumPermission.DoesNotExist:
                new_perm = GroupForumPermission(
                    permission=perm.permission, forum=forum_to,
                    group=perm.group,
                )
            new_perm.has_perm = perm.has_perm
            new_perm.save()


admin.site.register(Forum, ForumAdmin)



"""
    Forum member shortcuts
    ======================

    This module defines shortcut functions allowing to easily get forum member information.

"""

from machina.conf import settings as machina_settings


def get_forum_member_display_name(user):
    """ Given a specific user, returns their related display name. """
    return getattr(user, machina_settings.USER_DISPLAY_NAME_METHOD)()



"""
    Forum permission model admin definitions
    ========================================

    This module defines admin classes used to populate the Django administration dashboard.

"""

from django.contrib import admin

from machina.core.db.models import get_model


ForumPermission = get_model('forum_permission', 'ForumPermission')
GroupForumPermission = get_model('forum_permission', 'GroupForumPermission')
UserForumPermission = get_model('forum_permission', 'UserForumPermission')


class ForumPermissionAdmin(admin.ModelAdmin):
    """ The Forum Permission model admin. """

    search_fields = ('codename', )
    list_display = ('name', 'codename', )


class GroupForumPermissionAdmin(admin.ModelAdmin):
    """ The Group Forum Permission model admin. """

    search_fields = ('permission__codename', 'group__name', )
    list_display = ('group', 'forum', 'permission', 'has_perm', )
    list_editables = ('has_perm', )
    raw_id_fields = ('group', )
    list_filter = ['forum', 'group']


class UserForumPermissionAdmin(admin.ModelAdmin):
    """ The User Forum Permission model admin. """

    search_fields = ('permission__codename', 'user__username', )
    list_display = (
        'user',
        'anonymous_user',
        'authenticated_user',
        'forum',
        'permission',
        'has_perm',
    )
    list_editables = ('has_perm', )
    raw_id_fields = ('user', )
    list_filter = ['forum']


admin.site.register(ForumPermission, ForumPermissionAdmin)
admin.site.register(GroupForumPermission, GroupForumPermissionAdmin)
admin.site.register(UserForumPermission, UserForumPermissionAdmin)



"""
    Forum conversation signals
    ==========================

    This modules defines Django signals that can be triggered by the ``forum`` application.

"""

import django.dispatch


# "topic", "user", "request", "response"
topic_viewed = django.dispatch.Signal()



"""
    Forum polls model admin definitions
    ===================================

    This module defines admin classes used to populate the Django administration dashboard.

"""

from django.contrib import admin

from machina.core.db.models import get_model


TopicPoll = get_model('forum_polls', 'TopicPoll')
TopicPollOption = get_model('forum_polls', 'TopicPollOption')
TopicPollVote = get_model('forum_polls', 'TopicPollVote')


class TopicPollOptionInline(admin.TabularInline):
    model = TopicPollOption
    extra = 1


class TopicPollOptionAdmin(admin.ModelAdmin):
    """ The Topic Poll Option model admin. """

    list_display = ('__str__', 'poll', 'text',)
    search_fields = ('text',)


class TopicPollAdmin(admin.ModelAdmin):
    """ The Topic Poll model admin. """

    inlines = (TopicPollOptionInline,)
    list_display = ('topic', 'duration', 'max_options', 'user_changes', 'hide_results')
    list_filter = ('created', 'updated',)
    search_fields = ('topic__subject',)


class TopicPollVoteAdmin(admin.ModelAdmin):
    """ The Topic Poll Vote model admin. """

    list_display = ('__str__', 'voter',)


admin.site.register(TopicPoll, TopicPollAdmin)
admin.site.register(TopicPollOption, TopicPollOptionAdmin)
admin.site.register(TopicPollVote, TopicPollVoteAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2017-09-29 04:27
from __future__ import unicode_literals

from django.db import migrations


def update_forum_last_topic_and_last_post(apps, schema_editor):
    Forum = apps.get_model('forum', 'Forum')
    Topic = apps.get_model('forum_conversation', 'Topic')
    for forum in Forum.objects.all():
        topics = Topic.objects.filter(forum=forum, approved=True)
        if topics.exists():
            forum.last_post_id = topics[0].last_post_id
            forum.save()


class Migration(migrations.Migration):

    dependencies = [
        ('forum', '0008_forum_last_post'),
    ]

    operations = [
        migrations.RunPython(
            update_forum_last_topic_and_last_post, reverse_code=migrations.RunPython.noop),
    ]



"""
    Forum search indexes
    ====================

    This module defines search indexes allowing to perform searches among forum topics and posts.

"""

from haystack import indexes

from machina.core.db.models import get_model
from machina.core.loading import get_class


Post = get_model('forum_conversation', 'Post')

get_forum_member_display_name = get_class('forum_member.shortcuts', 'get_forum_member_display_name')


class PostIndex(indexes.SearchIndex, indexes.Indexable):
    """ Defines the data stored in the Post indexes. """

    text = indexes.CharField(
        document=True, use_template=True, template_name='forum_search/post_text.txt',
    )

    poster = indexes.IntegerField(model_attr='poster_id', null=True)
    poster_name = indexes.CharField()

    forum = indexes.IntegerField(model_attr='topic__forum_id')
    forum_slug = indexes.CharField()
    forum_name = indexes.CharField()

    topic = indexes.IntegerField(model_attr='topic_id')
    topic_slug = indexes.CharField()
    topic_subject = indexes.CharField()

    created = indexes.DateTimeField(model_attr='created')
    updated = indexes.DateTimeField(model_attr='updated')

    def get_model(self):
        return Post

    def prepare_poster_name(self, obj):
        return get_forum_member_display_name(obj.poster) if obj.poster else obj.username

    def prepare_forum_slug(self, obj):
        return obj.topic.forum.slug

    def prepare_forum_name(self, obj):
        return obj.topic.forum.name

    def prepare_topic_slug(self, obj):
        return obj.topic.slug

    def prepare_topic_subject(self, obj):
        return obj.topic.subject

    def index_queryset(self, using=None):
        return Post.objects.all().exclude(approved=False)

    def read_queryset(self, using=None):
        return Post.objects.all().exclude(approved=False).select_related('topic', 'poster')



"""
    Forum URLs
    ==========

    This module defines URL patterns associated with the django-machina's ``forum`` application.

"""

from django.urls import path

from machina.core.loading import get_class
from machina.core.urls import URLPatternsFactory


class ForumURLPatternsFactory(URLPatternsFactory):
    """ Allows to generate the URL patterns of the ``forum`` application. """

    app_namespace = 'forum'

    index_view = get_class('forum.views', 'IndexView')
    forum_view = get_class('forum.views', 'ForumView')

    def get_urlpatterns(self):
        """ Returns the URL patterns managed by the considered factory / application. """
        return [
            path('', self.index_view.as_view(), name='index'),
            path(
                'forum/<str:slug>-<int:pk>/',
                self.forum_view.as_view(),
                name='forum',
            ),
        ]


urlpatterns_factory = ForumURLPatternsFactory()



"""
    Forum moderation views
    ======================

    This module defines views provided by the ``forum_moderation`` application.

"""

from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import DeleteView, DetailView, ListView
from django.views.generic.detail import (
    BaseDetailView, SingleObjectMixin, SingleObjectTemplateResponseMixin
)
from django.views.generic.edit import FormMixin, ProcessFormView

from machina.conf import settings as machina_settings
from machina.core.db.models import get_model
from machina.core.loading import get_class


Forum = get_model('forum', 'Forum')
Post = get_model('forum_conversation', 'Post')
Topic = get_model('forum_conversation', 'Topic')

TopicMoveForm = get_class('forum_moderation.forms', 'TopicMoveForm')

PermissionRequiredMixin = get_class('forum_permission.viewmixins', 'PermissionRequiredMixin')


class TopicLockView(PermissionRequiredMixin, SingleObjectTemplateResponseMixin, BaseDetailView):
    """ Provides the ability to lock forum topics. """

    context_object_name = 'topic'
    model = Topic
    success_message = _('This topic has been locked successfully.')
    template_name = 'forum_moderation/topic_lock.html'

    def lock(self, request, *args, **kwargs):
        """ Locks the considered topic and retirects the user to the success URL. """
        self.object = self.get_object()
        success_url = self.get_success_url()
        self.object.status = Topic.TOPIC_LOCKED
        self.object.save()
        messages.success(self.request, self.success_message)
        return HttpResponseRedirect(success_url)

    def post(self, request, *args, **kwargs):
        """ Handles POST requests. """
        return self.lock(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)
        topic = self.get_object()
        context['forum'] = topic.forum
        return context

    def get_success_url(self):
        """ Returns the success URL to redirect the user to. """
        return reverse(
            'forum_conversation:topic',
            kwargs={
                'forum_slug': self.object.forum.slug,
                'forum_pk': self.object.forum.pk,
                'slug': self.object.slug,
                'pk': self.object.pk,
            },
        )

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_object().forum

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permission check. """
        return self.request.forum_permission_handler.can_lock_topics(obj, user)


class TopicUnlockView(PermissionRequiredMixin, SingleObjectTemplateResponseMixin, BaseDetailView):
    """ Provides the ability to unlock forum topics. """

    context_object_name = 'topic'
    model = Topic
    template_name = 'forum_moderation/topic_unlock.html'
    success_message = _('This topic has been unlocked successfully.')

    def unlock(self, request, *args, **kwargs):
        """ Unlocks the considered topic and retirects the user to the success URL. """
        self.object = self.get_object()
        success_url = self.get_success_url()
        self.object.status = Topic.TOPIC_UNLOCKED
        self.object.save()
        messages.success(self.request, self.success_message)
        return HttpResponseRedirect(success_url)

    def post(self, request, *args, **kwargs):
        """ Handles POST requests. """
        return self.unlock(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)
        topic = self.get_object()
        context['forum'] = topic.forum
        return context

    def get_success_url(self):
        """ Returns the success URL to redirect the user to. """
        return reverse(
            'forum_conversation:topic',
            kwargs={
                'forum_slug': self.object.forum.slug,
                'forum_pk': self.object.forum.pk,
                'slug': self.object.slug,
                'pk': self.object.pk,
            },
        )

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_object().forum

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permissions check. """
        return self.request.forum_permission_handler.can_lock_topics(obj, user)


class TopicDeleteView(PermissionRequiredMixin, DeleteView):
    """ Provides the ability to delete forum topics. """

    context_object_name = 'topic'
    model = Topic
    success_message = _('This topic has been deleted successfully.')
    template_name = 'forum_moderation/topic_delete.html'

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)
        topic = self.get_object()
        context['forum'] = topic.forum
        return context

    def get_success_url(self):
        """ Returns the success URL to redirect the user to. """
        messages.success(self.request, self.success_message)
        return reverse(
            'forum:forum', kwargs={'slug': self.object.forum.slug, 'pk': self.object.forum.pk},
        )

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_object().forum

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permissions check. """
        return self.request.forum_permission_handler.can_delete_topics(obj, user)


class TopicMoveView(
    PermissionRequiredMixin, SingleObjectTemplateResponseMixin, FormMixin, SingleObjectMixin,
    ProcessFormView,
):
    """ Provides the ability to move forum topics. """

    context_object_name = 'topic'
    form_class = TopicMoveForm
    model = Topic
    success_message = _('This topic has been moved successfully.')
    template_name = 'forum_moderation/topic_move.html'

    def get(self, request, *args, **kwargs):
        """ Handles GET requests. """
        self.object = self.get_object()
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        """ Handles POST requests. """
        self.object = self.get_object()
        return super().post(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)
        topic = self.get_object()
        context['forum'] = topic.forum
        return context

    def get_form_kwargs(self):
        """ Returns the keyword arguments used to initialize the associated form. """
        kwargs = super().get_form_kwargs()
        kwargs.update({
            'topic': self.object,
            'user': self.request.user,
        })
        return kwargs

    def form_valid(self, form):
        """ Handles a valid form. """
        # Move the topic
        topic = self.object
        new_forum = form.cleaned_data['forum']
        topic.forum = new_forum

        # Eventually lock the topic
        if form.cleaned_data['lock_topic']:
            topic.status = Topic.TOPIC_LOCKED
        else:
            topic.status = Topic.TOPIC_MOVED

        topic.save()

        messages.success(self.request, self.success_message)
        return HttpResponseRedirect(self.get_success_url())

    def get_success_url(self):
        """ Returns the success URL to redirect the user to. """
        return reverse(
            'forum_conversation:topic',
            kwargs={
                'forum_slug': self.object.forum.slug,
                'forum_pk': self.object.forum.pk,
                'slug': self.object.slug,
                'pk': self.object.pk,
            },
        )

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_object().forum

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permissions check. """
        return self.request.forum_permission_handler.can_move_topics(obj, user)


class TopicUpdateTypeBaseView(
    PermissionRequiredMixin, SingleObjectTemplateResponseMixin, BaseDetailView,
):
    """ Provides the ability to change the type of forum topics: normal, sticky topic or announce.
    """

    context_object_name = 'topic'
    model = Topic
    success_message = _('This topic type has been changed successfully.')
    template_name = 'forum_moderation/topic_update_type.html'

    # The following attributes should be defined in subclasses.
    question = ''
    target_type = None

    def update_type(self, request, *args, **kwargs):
        """ Updates the type of the considered topic and retirects the user to the success URL.
        """
        self.object = self.get_object()
        success_url = self.get_success_url()
        self.object.type = self.target_type
        self.object.save()
        messages.success(self.request, self.success_message)
        return HttpResponseRedirect(success_url)

    def post(self, request, *args, **kwargs):
        """ Handles POST requests. """
        return self.update_type(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)
        context['question'] = self.question
        topic = self.get_object()
        context['forum'] = topic.forum
        return context

    def get_success_url(self):
        """ Returns the success URL to redirect the user to. """
        return reverse(
            'forum_conversation:topic', kwargs={
                'forum_slug': self.object.forum.slug,
                'forum_pk': self.object.forum.pk,
                'slug': self.object.slug,
                'pk': self.object.pk,
            },
        )

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_object().forum


class TopicUpdateToNormalTopicView(TopicUpdateTypeBaseView):
    """ Provides the ability to switch a topic to a normal topic. """

    question = _('Would you want to change this topic to a default topic?')
    target_type = Topic.TOPIC_POST

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permissions check. """
        return self.request.forum_permission_handler.can_update_topics_to_normal_topics(obj, user)


class TopicUpdateToStickyTopicView(TopicUpdateTypeBaseView):
    """ Provides the ability to switch a topic to a sticky topic. """

    question = _('Would you want to change this topic to a sticky topic?')
    target_type = Topic.TOPIC_STICKY

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permissions check. """
        return self.request.forum_permission_handler.can_update_topics_to_sticky_topics(obj, user)


class TopicUpdateToAnnounceView(TopicUpdateTypeBaseView):
    """ Provides the ability to switch a topic to an announce. """

    question = _('Would you want to change this topic to an announce?')
    target_type = Topic.TOPIC_ANNOUNCE

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permissions check. """
        return self.request.forum_permission_handler.can_update_topics_to_announces(obj, user)


class ModerationQueueListView(PermissionRequiredMixin, ListView):
    """ Displays the moderation queue. """

    context_object_name = 'posts'
    model = Post
    paginate_by = machina_settings.TOPIC_POSTS_NUMBER_PER_PAGE
    template_name = 'forum_moderation/moderation_queue/list.html'

    def get_queryset(self):
        """ Returns the list of items for this view. """
        forums = self.request.forum_permission_handler.get_moderation_queue_forums(
            self.request.user,
        )
        qs = super().get_queryset()
        qs = qs.filter(topic__forum__in=forums, approved=False)
        return qs.order_by('-created')

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permissions check. """
        return self.request.forum_permission_handler.can_access_moderation_queue(user)


class ModerationQueueDetailView(PermissionRequiredMixin, DetailView):
    """ Displays the details of an item in the moderation queue. """

    context_object_name = 'post'
    model = Post
    template_name = 'forum_moderation/moderation_queue/detail.html'

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)

        post = self.object
        topic = post.topic

        # Handles the case when a poll is associated to the topic.
        try:
            if hasattr(topic, 'poll') and topic.poll.options.exists():
                poll = topic.poll
                context['poll'] = poll
                context['poll_options'] = poll.options.all()
        except ObjectDoesNotExist:  # pragma: no cover
            pass

        if not post.is_topic_head:
            # Add the topic review
            previous_posts = (
                topic.posts
                .filter(approved=True, created__lte=post.created)
                .select_related('poster', 'updated_by')
                .prefetch_related('attachments', 'poster__forum_profile')
                .order_by('-created')
            )
            previous_posts = previous_posts[:machina_settings.TOPIC_REVIEW_POSTS_NUMBER]
            context['previous_posts'] = previous_posts

        return context

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_object().topic.forum

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permissions check. """
        return self.request.forum_permission_handler.can_approve_posts(obj, user)


class PostApproveView(PermissionRequiredMixin, SingleObjectTemplateResponseMixin, BaseDetailView):
    """ Provides the ability to approve queued forum posts. """

    context_object_name = 'post'
    model = Post
    success_message = _('This post has been approved successfully.')
    template_name = 'forum_moderation/moderation_queue/post_approve.html'

    def approve(self, request, *args, **kwargs):
        """ Approves the considered post and retirects the user to the success URL. """
        self.object = self.get_object()
        success_url = self.get_success_url()
        self.object.approved = True
        self.object.save()
        messages.success(self.request, self.success_message)
        return HttpResponseRedirect(success_url)

    def post(self, request, *args, **kwargs):
        """ Handles POST requests. """
        return self.approve(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)
        context['forum'] = self.get_object().topic.forum
        return context

    def get_success_url(self):
        """ Returns the success URL to redirect the user to. """
        return reverse('forum_moderation:queue')

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_object().topic.forum

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permissions check. """
        return self.request.forum_permission_handler.can_approve_posts(obj, user)


class PostDisapproveView(
    PermissionRequiredMixin, SingleObjectTemplateResponseMixin, BaseDetailView,
):
    """ Provides the ability to disapprove queued forum posts. """

    context_object_name = 'post'
    model = Post
    success_message = _('This post has been disapproved successfully.')
    template_name = 'forum_moderation/moderation_queue/post_disapprove.html'

    def disapprove(self, request, *args, **kwargs):
        """ Disapproves the considered post and retirects the user to the success URL. """
        self.object = self.get_object()
        success_url = self.get_success_url()
        self.object.delete()
        messages.success(self.request, self.success_message)
        return HttpResponseRedirect(success_url)

    def post(self, request, *args, **kwargs):
        """ Handles POST requests. """
        return self.disapprove(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)
        context['forum'] = self.get_object().topic.forum
        return context

    def get_success_url(self):
        """ Returns the success URL to redirect the user to. """
        return reverse('forum_moderation:queue')

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_object().topic.forum

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permissions check. """
        return self.request.forum_permission_handler.can_approve_posts(obj, user)



# Generated by Django 2.1.2 on 2018-11-03 19:01

from django.db import migrations
import machina.apps.forum.abstract_models
import machina.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('forum', '0009_auto_20170928_2327'),
    ]

    operations = [
        migrations.AlterField(
            model_name='forum',
            name='image',
            field=machina.models.fields.ExtendedImageField(blank=True, null=True, upload_to=machina.apps.forum.abstract_models.get_forum_image_upload_to, verbose_name='Forum image'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('forum_conversation', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='post',
            name='anonymous_key',
            field=models.CharField(max_length=100, null=True, verbose_name='Anonymous user forum key', blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-05-05 02:08
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('forum', '0003_remove_forum_is_active'),
    ]

    operations = [
        migrations.AddField(
            model_name='forum',
            name='direct_posts_count',
            field=models.PositiveIntegerField(blank=True, default=0, editable=False, verbose_name='Direct number of posts'),
        ),
        migrations.AddField(
            model_name='forum',
            name='direct_topics_count',
            field=models.PositiveIntegerField(blank=True, default=0, editable=False, verbose_name='Direct number of topics'),
        ),
    ]



# Generated by Django 2.1.7 on 2019-03-20 03:40

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('forum_permission', '0003_remove_forumpermission_name'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='forumpermission',
            name='is_global',
        ),
        migrations.RemoveField(
            model_name='forumpermission',
            name='is_local',
        ),
    ]



"""
    Forum search views
    ==================

    This module defines views provided by the ``forum_search`` application.

"""

from haystack import views


class FacetedSearchView(views.FacetedSearchView):
    """ Allows to search within forums. """

    template = 'forum_search/search.html'

    def build_form(self):
        form = super().build_form(form_kwargs={'user': self.request.user})
        return form



"""
    Forum conversation signal receivers
    ===================================

    This module defines signal receivers.

"""

from django.db.models import F
from django.dispatch import receiver

from machina.apps.forum_conversation.signals import topic_viewed


@receiver(topic_viewed)
def update_topic_counter(sender, topic, user, request, response, **kwargs):
    """ Handles the update of the views counter associated with topics. """
    topic.__class__._default_manager.filter(id=topic.id).update(views_count=F('views_count') + 1)



"""
    Forum tracking models
    =====================

    This module defines models provided by the ``forum_tracking`` application.

"""

from machina.apps.forum_tracking.abstract_models import (
    AbstractForumReadTrack, AbstractTopicReadTrack
)
from machina.core.db.models import model_factory


ForumReadTrack = model_factory(AbstractForumReadTrack)
TopicReadTrack = model_factory(AbstractTopicReadTrack)



"""
    Forum signals
    =============

    This modules defines Django signals that can be triggered by the ``forum`` application.

"""

import django.dispatch


# Arguments:"previous_parent"
forum_moved = django.dispatch.Signal()
# Arguments:"forum", "user", "request", "response"
forum_viewed = django.dispatch.Signal()



import sys
import traceback

from django.conf import settings


class AppNotFoundError(Exception):
    pass


class ClassNotFoundError(Exception):
    pass


def get_class(module_label, classname):
    return get_classes(module_label, [classname, ])[0]


def get_classes(module_label, classnames):
    """ Imports a set of classes from a given module.

    Usage::

        get_classes('forum.models', ['Forum', 'ForumReadTrack', ])

    """
    app_label = module_label.split('.')[0]

    app_module_path = _get_app_module_path(module_label)
    if not app_module_path:
        raise AppNotFoundError('No app found matching \'{}\''.format(module_label))

    # Determines the full module path by appending the module label
    # to the base package path of the considered application.
    module_path = app_module_path
    if '.' in app_module_path:
        base_package = app_module_path.rsplit('.' + app_label, 1)[0]
        module_path = '{}.{}'.format(base_package, module_label)

    # Try to import this module from the related app that is specified
    # in the Django settings.
    local_imported_module = _import_module(module_path, classnames)

    # If the module we tried to import is not located inside the machina
    # vanilla apps, try to import it from the corresponding machina app.
    machina_imported_module = None
    if not app_module_path.startswith('machina.apps'):
        machina_imported_module = _import_module(
            '{}.{}'.format('machina.apps', module_label), classnames,
        )

    if local_imported_module is None and machina_imported_module is None:
        raise AppNotFoundError('Error importing \'{}\''.format(module_path))

    # Any local module is prioritized over the corresponding machina module
    imported_modules = [
        m for m in (local_imported_module, machina_imported_module) if m is not None
    ]
    return _pick_up_classes(imported_modules, classnames)


def _import_module(module_path, classnames):
    """ Tries to import the given Python module path. """
    try:
        imported_module = __import__(module_path, fromlist=classnames)
        return imported_module
    except ImportError:
        # In case of an ImportError, the module being loaded generally does not exist. But an
        # ImportError can occur if the module being loaded exists and another import located inside
        # it failed.
        #
        # In order to provide a meaningfull traceback, the execution information can be inspected in
        # order to determine which case to consider. If the execution information provides more than
        # a certain amount of frames, this means that an ImportError occured while loading the
        # initial Python module.
        __, __, exc_traceback = sys.exc_info()
        frames = traceback.extract_tb(exc_traceback)
        if len(frames) > 1:
            raise


def _pick_up_classes(modules, classnames):
    """ Given a list of class names to retrieve, try to fetch them from the specified list of
        modules and returns the list of the fetched classes.
    """
    klasses = []
    for classname in classnames:
        klass = None
        for module in modules:
            if hasattr(module, classname):
                klass = getattr(module, classname)
                break
        if not klass:
            raise ClassNotFoundError('Error fetching \'{}\' in {}'.format(
                classname, str([module.__name__ for module in modules]))
            )
        klasses.append(klass)
    return klasses


def _get_app_module_path(module_label):
    """ Given a module label, loop over the apps specified in the INSTALLED_APPS to find the
        corresponding application module path.
    """
    app_name = module_label.rsplit('.', 1)[0]
    for app in settings.INSTALLED_APPS:
        if app.endswith('.' + app_name) or app == app_name:
            return app
        path = app.split('.')
        if len(path) >= 3 and path[-2] == 'apps' and path[-3] == app_name:
            return app
    return None



"""
    Forum permission models
    =======================

    This module defines models provided by the ``forum_permission`` application.

"""

from machina.apps.forum_permission.abstract_models import (
    AbstractForumPermission, AbstractGroupForumPermission, AbstractUserForumPermission
)
from machina.core.db.models import model_factory


ForumPermission = model_factory(AbstractForumPermission)
GroupForumPermission = model_factory(AbstractGroupForumPermission)
UserForumPermission = model_factory(AbstractUserForumPermission)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.2 on 2016-02-28 19:51
from __future__ import unicode_literals

from django.db import migrations
import machina.core.validators
import machina.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('forum_conversation', '0002_post_anonymous_key'),
    ]

    operations = [
        migrations.AlterField(
            model_name='post',
            name='content',
            field=machina.models.fields.MarkupTextField(no_rendered_field=True, validators=[machina.core.validators.NullableMaxLengthValidator(None)], verbose_name='Content'),
        ),
    ]



import factory
import factory.django

from machina.core.db.models import get_model


ForumPermission = get_model('forum_permission', 'ForumPermission')
GroupForumPermission = get_model('forum_permission', 'GroupForumPermission')
UserForumPermission = get_model('forum_permission', 'UserForumPermission')


class ForumPermissionFactory(factory.django.DjangoModelFactory):
    codename = factory.Sequence(lambda n: 'perm-{}'.format(n))

    class Meta:
        model = ForumPermission


class UserForumPermissionFactory(factory.django.DjangoModelFactory):
    permission = factory.SubFactory(ForumPermissionFactory)

    class Meta:
        model = UserForumPermission


class GroupForumPermissionFactory(factory.django.DjangoModelFactory):
    permission = factory.SubFactory(ForumPermissionFactory)

    class Meta:
        model = GroupForumPermission



"""
    Forum member signal receivers
    =============================

    This module defines signal receivers.

"""

from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import F
from django.db.models.signals import post_delete, pre_save
from django.dispatch import receiver

from machina.core.db.models import get_model


User = get_user_model()

Post = get_model('forum_conversation', 'Post')
ForumProfile = get_model('forum_member', 'ForumProfile')


@receiver(pre_save, sender=Post)
def increase_posts_count(sender, instance, **kwargs):
    """ Increases the member's post count after a post save.

    This receiver handles the update of the profile related to the user who is the poster of the
    forum post being created or updated.
    """

    if kwargs.get('raw'):
        # do nothing, when loading data (fixtures)
        return

    if instance.poster is None:
        # An anonymous post is considered. No profile can be updated in
        # that case.
        return

    profile, dummy = ForumProfile.objects.get_or_create(user=instance.poster)
    increase_posts_count = False

    if instance.pk:
        try:
            old_instance = instance.__class__._default_manager.get(pk=instance.pk)
        except ObjectDoesNotExist:  # pragma: no cover
            # This should never happen (except with django loaddata command)
            increase_posts_count = True
            old_instance = None
        if old_instance and old_instance.approved is False and instance.approved is True:
            increase_posts_count = True
    elif instance.approved:
        increase_posts_count = True

    if increase_posts_count:
        profile.posts_count = F('posts_count') + 1
        profile.save()


@receiver(pre_save, sender=Post)
def decrease_posts_count_after_post_unaproval(sender, instance, **kwargs):
    """ Decreases the member's post count after a post unaproval.

    This receiver handles the unaproval of a forum post: the posts count associated with the post's
    author is decreased.
    """

    if kwargs.get('raw'):
        # do nothing, when loading data (fixtures)
        return

    if not instance.pk or not instance.poster:
        # Do not consider posts being created or posts of anonymous users
        return

    profile, dummy = ForumProfile.objects.get_or_create(user=instance.poster)

    try:
        old_instance = instance.__class__._default_manager.get(pk=instance.pk)
    except ObjectDoesNotExist:  # pragma: no cover
        # This should never happen (except with django loaddata command)
        return

    if old_instance and old_instance.approved is True and instance.approved is False:
        profile.posts_count = F('posts_count') - 1
        profile.save()


@receiver(post_delete, sender=Post)
def decrease_posts_count_after_post_deletion(sender, instance, **kwargs):
    """ Decreases the member's post count after a post deletion.

    This receiver handles the deletion of a forum post: the posts count related to the post's
    author is decreased.
    """
    if not instance.approved:
        # If a post has not been approved, it has not been counted.
        # So do not decrement count
        return

    try:
        assert instance.poster_id is not None
        poster = User.objects.get(pk=instance.poster_id)
    except AssertionError:
        # An anonymous post is considered. No profile can be updated in
        # that case.
        return
    except ObjectDoesNotExist:  # pragma: no cover
        # This can happen if a User instance is deleted. In that case the
        # User instance is not available and the receiver should return.
        return

    profile, dummy = ForumProfile.objects.get_or_create(user=poster)
    if profile.posts_count:
        profile.posts_count = F('posts_count') - 1
        profile.save()



"""
    Forum search URLs
    =================

    This module defines URL patterns associated with the django-machina's ``forum_search``
    application.

"""

from django.urls import path
from haystack.views import search_view_factory

from machina.core.loading import get_class
from machina.core.urls import URLPatternsFactory


class ForumSearchURLPatternsFactory(URLPatternsFactory):
    """ Allows to generate the URL patterns of the ``forum_search`` application. """

    app_namespace = 'forum_search'

    search_view = get_class('forum_search.views', 'FacetedSearchView')
    search_form = get_class('forum_search.forms', 'SearchForm')

    def get_urlpatterns(self):
        """ Returns the URL patterns managed by the considered factory / application. """
        return [
            path(
                '',
                search_view_factory(view_class=self.search_view, form_class=self.search_form),
                name='search',
            ),
        ]


urlpatterns_factory = ForumSearchURLPatternsFactory()



"""
    Forum member URLs
    =================

    This module defines URL patterns associated with the django-machina's ``forum_member``
    application.

"""

from django.urls import path

from machina.core.loading import get_class
from machina.core.urls import URLPatternsFactory


class ForumMemberURLPatternsFactory(URLPatternsFactory):
    """ Allows to generate the URL patterns of the ``forum_member`` application. """

    app_namespace = 'forum_member'

    user_posts_list = get_class('forum_member.views', 'UserPostsView')
    forum_profile_detail_view = get_class('forum_member.views', 'ForumProfileDetailView')
    forum_profile_update_view = get_class('forum_member.views', 'ForumProfileUpdateView')
    topic_subscribe_view = get_class('forum_member.views', 'TopicSubscribeView')
    topic_unsubscribe_view = get_class('forum_member.views', 'TopicUnsubscribeView')
    topic_subscription_list_view = get_class('forum_member.views', 'TopicSubscriptionListView')

    def get_urlpatterns(self):
        """ Returns the URL patterns managed by the considered factory / application. """
        return [
            path(
                'profile/edit/',
                self.forum_profile_update_view.as_view(),
                name='profile_update',
            ),
            path(
                'profile/<str:pk>/',
                self.forum_profile_detail_view.as_view(),
                name='profile',
            ),
            path(
                'profile/<str:pk>/posts/',
                self.user_posts_list.as_view(),
                name='user_posts',
            ),
            path(
                'subscriptions/',
                self.topic_subscription_list_view.as_view(),
                name='user_subscriptions',
            ),
            path(
                'topic/<int:pk>/subscribe/',
                self.topic_subscribe_view.as_view(),
                name='topic_subscribe',
            ),
            path(
                'topic/<int:pk>/unsubscribe/',
                self.topic_unsubscribe_view.as_view(),
                name='topic_unsubscribe',
            ),
        ]


urlpatterns_factory = ForumMemberURLPatternsFactory()



"""
    Forum permission checker
    ========================

    This module defines a ``ForumPermissionChecker`` abstraction that allows to check forum
    permissions on specific forum instances.

"""

from collections import OrderedDict

from django.contrib.auth import get_user_model
from django.db.models import Q

from machina.conf import settings as machina_settings
from machina.core.db.models import get_model


ForumPermission = get_model('forum_permission', 'ForumPermission')
GroupForumPermission = get_model('forum_permission', 'GroupForumPermission')
UserForumPermission = get_model('forum_permission', 'UserForumPermission')


class ForumPermissionChecker:
    """ The ForumPermissionChecker allows to check forum permissions on Forum instances. """

    def __init__(self, user):
        self.user = user
        self._forum_perms_cache = {}

    def has_perm(self, perm, forum):
        """ Checks if the considered user has given permission for the passed forum. """
        if not self.user.is_anonymous and not self.user.is_active:
            # An inactive user cannot have permissions
            return False
        elif self.user and self.user.is_superuser:
            # The superuser have all permissions
            return True
        return perm in self.get_perms(forum)

    def get_perms(self, forum):
        """ Returns the list of permission codenames of all permissions for the given forum. """
        # An inactive user has no permissions.
        if not self.user.is_anonymous and not self.user.is_active:
            return []

        if forum is None:
            forum_identifier = 'global'
        else:
            forum_identifier = forum.id

        if forum_identifier not in self._forum_perms_cache:
            if self.user and self.user.is_superuser:
                # The superuser has all the permissions.
                permcodes = list(ForumPermission.objects.values_list('codename', flat=True))
            elif self.user:
                perms = self.get_perms_for_forumlist([forum], None)
                permcodes = perms[forum]

            self._forum_perms_cache[forum_identifier] = permcodes

        return self._forum_perms_cache[forum_identifier]

    def get_perms_for_forumlist(self, forums, perm_codenames=None):
        """
            Computes and returns a dictionary of [forum] to (set of permissions) for the user,
            taking into account precendence of permissions:
                - forum > global
                - user > group > all_authenticated_users
            Expects:
            - forums to be a list of forum objects
            - perm_codenames to be a list of permission codes (strings) to look for or None
        """
        globally_granted_all_users_perms = set()
        globally_granted_group_perms, globally_nongranted_group_perms = set(), set()

        user_perms = (
            UserForumPermission.objects.select_related()
            .filter(Q(forum__isnull=True) | Q(forum__in=forums))
        )
        if perm_codenames:
            user_perms = user_perms.filter(permission__codename__in=perm_codenames)

        # Do some additional filtering on (type of) user
        if self.user.is_anonymous:
            user_perms = user_perms.filter(anonymous_user=True)
            all_users_perms = None
            group_perms = None
        else:
            two_types_user_perms = user_perms.filter(Q(authenticated_user=True) | Q(user=self.user))
            all_users_perms = [p for p in two_types_user_perms if p.authenticated_user]
            user_perms = [p for p in two_types_user_perms if p.user]
            # Now get group permissions
            user_model = get_user_model()
            user_groups_related_name = user_model.groups.field.related_query_name()
            group_perms = (
                GroupForumPermission.objects.select_related()
                .filter(**{'group__{}'.format(user_groups_related_name): self.user})
                .filter(Q(forum__isnull=True) | Q(forum__in=forums))
            )
            if perm_codenames:
                group_perms = group_perms.filter(permission__codename__in=perm_codenames)

            # The following 3 lists can already be made outside of the loop over forums
            # But only for non-anonymous users so we do it here
            globally_granted_all_users_perms = list(
                filter(lambda p: p.has_perm and p.forum_id is None, all_users_perms)
            )
            globally_granted_group_perms = list(
                filter(lambda p: p.has_perm and p.forum_id is None, group_perms)
            )
            globally_nongranted_group_perms = list(
                filter(lambda p: not p.has_perm and p.forum_id is None, group_perms)
            )

        # Computes the list of permissions that are non-granted for all the forums.
        globally_nongranted_user_perms = list(
            filter(lambda p: not p.has_perm and p.forum_id is None, user_perms)
        )
        # Computes the list of permissions that are granted for all the forums.
        globally_granted_user_perms = list(
            filter(lambda p: p.has_perm and p.forum_id is None, user_perms)
        )
        globally_granted_user_permcodes = [
            p.permission.codename for p in globally_granted_user_perms
        ]
        # If the considered user has no global permissions, the permissions defined by the
        # DEFAULT_AUTHENTICATED_USER_FORUM_PERMISSIONS settings are used instead.
        if self.user.is_authenticated and not globally_granted_user_permcodes:
            globally_granted_user_permcodes = machina_settings.DEFAULT_AUTHENTICATED_USER_FORUM_PERMISSIONS  # noqa: E501

        forum_to_permissions = OrderedDict()
        for f in forums:
            """
            In every place that we check on f.id we first check if f is actually there
            It might be None because some tests only try global permissions without forum
            """
            # (Re)set these variables for each loop
            permcodes, granted_group_permcodes, granted_all_users_permcodes = set(), set(), set()

            # ########## BLOCK FOR PERMS SPECIFIC TO ONE LOGGED IN USER ###########
            per_forum_granted_user_perms = list(
                filter(lambda p: p.has_perm and f and p.forum_id == f.id, user_perms)
            )
            per_forum_granted_user_permcodes = [
                p.permission.codename for p in per_forum_granted_user_perms
            ]
            per_forum_nongranted_user_perms = list(
                filter(lambda p: not p.has_perm and f and p.forum_id == f.id, user_perms)
            )
            per_forum_nongranted_user_permcodes = [
                p.permission.codename for p in per_forum_nongranted_user_perms
            ]

            per_forum_granted_group_perms = []
            per_forum_nongranted_group_perms = []
            per_forum_nongranted_all_users_permcodes = []
            # ########## BLOCK FOR PERMS SPECIFIC TO GROUPS OF LOGGED IN USER ###########
            # If the user is a registered user, we have to check the permissions of its groups
            # in order to determine the additional permissions they could have.
            if not self.user.is_anonymous:
                if group_perms:
                    # A permission can be non-granted on user-forum level, and that takes
                    # precedence over granted group permissions so we do not add those to the list.
                    per_forum_granted_group_perms = list(
                        filter(lambda p: p.has_perm and f and p.forum_id == f.id and
                               p.permission_id not in
                               [q.permission_id for q in per_forum_nongranted_user_perms],
                               group_perms)
                    )
                    per_forum_granted_group_permcodes = [
                        p.permission.codename for p in per_forum_granted_group_perms
                    ]
                    # A permission can be granted on user-forum level, and that takes precedence
                    # over nongranted group permissions so we do not add those to the list.
                    per_forum_nongranted_group_perms = list(
                        filter(lambda p: not p.has_perm and f and p.forum_id == f.id and
                               p.permission_id not in
                               [q.permission_id for q in per_forum_granted_user_perms],
                               group_perms)
                    )

                    # Filter the globally granted group perms to those that were:
                    # - not set to non-granted on global-user level
                    # - and not set to non-granted on forum-group level
                    # - and not set to non-granted on forum-user level
                    globally_granted_group_perms = list(
                        filter(lambda p: p.has_perm and p.forum_id is None and
                               p.permission_id not in
                               [q.permission_id for q in globally_nongranted_user_perms] and
                               p.permission_id not in
                               [y.permission_id for y in per_forum_nongranted_group_perms] and
                               p.permission_id not in
                               [z.permission_id for z in per_forum_nongranted_user_perms],
                               globally_granted_group_perms)
                    )
                    globally_granted_group_permcodes = [
                        p.permission.codename for p in globally_granted_group_perms
                    ]

                    # Filter the globally non granted group perms to those that were:
                    # - not set to granted on global- user level
                    # - and not set to granted on forum-group level
                    # - and not set to granted on forum-user level
                    globally_nongranted_group_perms = list(
                        filter(lambda p: not p.has_perm and p.forum_id is None and
                               p.permission_id not in
                               [q.permission_id for q in globally_granted_user_perms] and
                               p.permission_id not in
                               [y.permission_id for y in per_forum_granted_group_perms] and
                               p.permission_id not in
                               [z.permission_id for z in per_forum_granted_user_perms],
                               globally_nongranted_group_perms)
                    )
                    granted_group_permcodes = set(globally_granted_group_permcodes +
                                                  per_forum_granted_group_permcodes)

                # ######### BLOCK FOR PERMS FOR EVERY LOGGED IN USER ##########
                # A permission can be non-granted on user-forum or group-forum level, and
                # that takes precedence over granted all_users permissions so we do not add
                # those to the list.
                per_forum_granted_all_users_perms = list(
                    filter(lambda p: p.has_perm and f and p.forum_id == f.id and
                           p.permission_id not in
                           [q.permission_id for q in per_forum_nongranted_user_perms] and
                           p.permission_id not in
                           [z.permission_id for z in per_forum_nongranted_group_perms],
                           all_users_perms)
                )
                per_forum_granted_all_users_permcodes = [
                    p.permission.codename for p in per_forum_granted_all_users_perms
                ]

                # A permission can be granted on user-forum or group-forum level, and that takes
                # precedence over nongranted all_user permissions so we do not add those to
                # the list
                per_forum_nongranted_all_users_perms = list(
                    filter(lambda p: not p.has_perm and f and p.forum_id == f.id and
                           p.permission_id not in
                           [q.permission_id for q in per_forum_granted_user_perms] and
                           p.permission_id not in
                           [z.permission_id for z in per_forum_granted_group_perms],
                           all_users_perms)
                )
                per_forum_nongranted_all_users_permcodes = [
                    p.permission.codename for p in per_forum_nongranted_all_users_perms
                ]
                # Filter the globally granted all users perms to those that were:
                # - not set to non-granted on forum-all_user level
                # - and not set to non-granted on global-user level
                # - and not set to non-granted on global-group level
                # - and not set to non-granted on forum-group level
                # - and not set to non-granted on forum-user level
                globally_granted_all_users_perms = list(
                    filter(lambda p: p.has_perm and p.forum_id is None and
                           p.permission_id not in
                           [y.permission_id for y in per_forum_nongranted_all_users_perms] and
                           p.permission_id not in
                           [q.permission_id for q in globally_nongranted_user_perms] and
                           p.permission_id not in
                           [a.permission_id for a in globally_nongranted_group_perms] and
                           p.permission_id not in
                           [x.permission_id for x in per_forum_nongranted_group_perms] and
                           p.permission_id not in
                           [z.permission_id for z in per_forum_nongranted_user_perms],
                           globally_granted_all_users_perms)
                )
                globally_granted_all_users_permcodes = [
                    p.permission.codename for p in globally_granted_all_users_perms
                ]
                granted_all_users_permcodes = set(globally_granted_all_users_permcodes +
                                                  per_forum_granted_all_users_permcodes)

            # Finally computes the list of permission codenames that are
            # granted to the user for the considered forum.
            # We can not do this earlier because
            # globally_granted_user_perms can be from the setting
            # DEFAULT_AUTHENTICATED_USER_FORUM_PERMISSIONS in which case
            # it only contains permission codes and not actual permission
            # objects that we can loop over and check against.
            granted_user_permcodes = [
                c for c in globally_granted_user_permcodes if
                c not in per_forum_nongranted_user_permcodes and
                c not in per_forum_nongranted_all_users_permcodes
            ]
            permcodes = set(granted_user_permcodes +
                            per_forum_granted_user_permcodes)
            # Includes the permissions granted for the user's groups and for all logged
            # in users (that were not overruled by more specific targets) in the initial
            # set of permission codenames.
            forum_to_permissions[f] = permcodes.union(granted_group_permcodes,
                                                      granted_all_users_permcodes)

        return forum_to_permissions



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-06-07 03:02
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('forum_tracking', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='forumreadtrack',
            name='mark_time',
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='topicreadtrack',
            name='mark_time',
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
    ]



"""
    Forum signal receivers
    ======================

    This module defines signal receivers.

"""

from django.db.models import F
from django.dispatch import receiver

from machina.apps.forum.signals import forum_viewed


@receiver(forum_viewed)
def update_forum_redirects_counter(sender, forum, user, request, response, **kwargs):
    """ Handles the update of the link redirects counter associated with link forums. """
    if forum.is_link and forum.link_redirects:
        forum.link_redirects_count = F('link_redirects_count') + 1
        forum.save()



import factory
import factory.django

from machina.core.db.models import get_model
from machina.test.factories.auth import UserFactory
from machina.test.factories.conversation import TopicFactory
from machina.test.factories.forum import ForumFactory


ForumReadTrack = get_model('forum_tracking', 'ForumReadTrack')
TopicReadTrack = get_model('forum_tracking', 'TopicReadTrack')


class ForumReadTrackFactory(factory.django.DjangoModelFactory):
    user = factory.SubFactory(UserFactory)
    forum = factory.SubFactory(ForumFactory)

    class Meta:
        model = ForumReadTrack


class TopicReadTrackFactory(factory.django.DjangoModelFactory):
    user = factory.SubFactory(UserFactory)
    topic = factory.SubFactory(TopicFactory)

    class Meta:
        model = TopicReadTrack



"""
    Forum polls forms
    =================

    This module defines forms provided by the ``forum_polls`` application.

"""

from django import forms
from django.core.exceptions import ValidationError
from django.forms.models import BaseModelFormSet, modelformset_factory
from django.utils.translation import gettext_lazy as _

from machina.conf import settings as machina_settings
from machina.core.db.models import get_model
from machina.core.shortcuts import get_object_or_none


TopicPoll = get_model('forum_polls', 'TopicPoll')
TopicPollOption = get_model('forum_polls', 'TopicPollOption')


class TopicPollOptionForm(forms.ModelForm):
    class Meta:
        model = TopicPollOption
        fields = ['text', ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Update the 'text' field
        self.fields['text'].label = ''
        self.fields['text'].widget.attrs['placeholder'] = _('Enter a poll option')
        self.fields['text'].required = False


class BaseTopicPollOptionFormset(BaseModelFormSet):
    topic = None
    poll = None

    def __init__(self, *args, **kwargs):
        self.topic = kwargs.pop('topic', None)
        if self.topic:
            self.poll = get_object_or_none(TopicPoll, topic=self.topic)

        super().__init__(*args, **kwargs)

        if self.poll is not None:
            for form in self.forms:
                form.instance.poll = self.poll

    def total_form_count(self):
        """
        This rewrite of total_form_count allows to add an empty form to the formset only when
        no initial data is provided.
        """
        total_forms = super().total_form_count()
        if not self.data and not self.files and self.initial_form_count() > 0:
            total_forms -= self.extra
        return total_forms

    def clean(self):
        if any(self.errors):
            return

        # At least two options must be defined
        number_of_options = 0
        for form in self.forms:
            if (
                not (
                    (self.can_delete and self._should_delete_form(form)) or
                    len(form.cleaned_data) == 0
                )
            ):
                number_of_options += 1
        if number_of_options < 2:
            raise forms.ValidationError('At least two poll options must be defined.')

    def save(self, commit=True, **kwargs):
        poll_question = kwargs.pop('poll_question', None)
        poll_max_options = kwargs.pop('poll_max_options', None)
        poll_duration = kwargs.pop('poll_duration', None)
        poll_user_changes = kwargs.pop('poll_user_changes', False)
        poll_hide_results = kwargs.pop('poll_hide_results', False)

        if self.poll is None:
            poll, _ = TopicPoll.objects.get_or_create(topic=self.topic)

            poll.question = poll_question
            poll.duration = poll_duration
            poll.max_options = poll_max_options
            poll.user_changes = poll_user_changes
            poll.hide_results = poll_hide_results
            poll.save()

            for form in self.forms:
                form.instance.poll = poll
        super().save(commit)


TopicPollOptionFormset = modelformset_factory(
    TopicPollOption, TopicPollOptionForm,
    formset=BaseTopicPollOptionFormset,
    can_delete=True, extra=2, max_num=machina_settings.POLL_MAX_OPTIONS_PER_POLL,
    validate_max=True,
)


class TopicPollVoteForm(forms.Form):
    def __init__(self, poll, *args, **kwargs):
        self.poll = poll
        super().__init__(*args, **kwargs)

        if poll.max_options == 1:
            self.fields['options'] = forms.ModelChoiceField(
                label='', queryset=poll.options.all(), empty_label=None,
                widget=forms.RadioSelect(),
            )
        else:
            self.fields['options'] = forms.ModelMultipleChoiceField(
                label='', queryset=poll.options.all(),
                widget=forms.CheckboxSelectMultiple(),
            )

    def clean_options(self):
        options = self.cleaned_data['options']
        if isinstance(options, TopicPollOption):
            options = [options, ]
        return options

    def clean(self):
        cleaned_data = super().clean()

        if 'options' not in cleaned_data:
            msg = _('You must specify an option when voting.')
            raise ValidationError(msg)

        options = cleaned_data['options']
        if len(options) > self.poll.max_options:
            msg = _('You have tried to vote for too many options.')
            raise ValidationError(msg)

        return cleaned_data



# Generated by Django 3.2.15 on 2022-09-11 20:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum_polls", "0002_auto_20151105_0029"),
    ]

    operations = [
        migrations.AddField(
            model_name="topicpoll",
            name="hide_results",
            field=models.BooleanField(default=False, verbose_name="Hide results"),
        ),
    ]



"""
    Forum attachments abstract models
    =================================

    This module defines abstract models provided by the ``forum_attachments`` application.

"""

import os

from django.db import models
from django.utils.translation import gettext_lazy as _

from machina.conf import settings as machina_settings


def get_attachment_file_upload_to(instance, filename):
    """ Returns a valid upload path for the file of an attachment. """
    return instance.get_file_upload_to(filename)


class AbstractAttachment(models.Model):
    """ Represents a post attachment. An attachment is always linked to a post. """

    post = models.ForeignKey(
        'forum_conversation.Post', related_name='attachments', on_delete=models.CASCADE,
        verbose_name=_('Post'),
    )
    file = models.FileField(upload_to=get_attachment_file_upload_to, verbose_name=_('File'))
    comment = models.CharField(max_length=255, verbose_name=_('Comment'), blank=True, null=True)

    class Meta:
        abstract = True
        app_label = 'forum_attachments'
        verbose_name = _('Attachment')
        verbose_name_plural = _('Attachments')

    def __str__(self):
        return '{}'.format(self.post.subject)

    @property
    def filename(self):
        """ Returns the filename of the considered attachment. """
        return os.path.basename(self.file.name)

    def get_file_upload_to(self, filename):
        """ Returns the path to upload the associated file to. """
        return os.path.join(machina_settings.ATTACHMENT_FILE_UPLOAD_TO, filename)



import factory
import factory.django
from faker import Faker

from machina.core.db.models import get_model
from machina.test.factories.auth import UserFactory
from machina.test.factories.conversation import TopicFactory


faker = Faker()

TopicPoll = get_model('forum_polls', 'TopicPoll')
TopicPollOption = get_model('forum_polls', 'TopicPollOption')
TopicPollVote = get_model('forum_polls', 'TopicPollVote')


class TopicPollFactory(factory.django.DjangoModelFactory):
    topic = factory.SubFactory(TopicFactory)
    question = faker.text(max_nb_chars=200)

    class Meta:
        model = TopicPoll


class TopicPollOptionFactory(factory.django.DjangoModelFactory):
    poll = factory.SubFactory(TopicPollFactory)
    text = faker.text(max_nb_chars=100)

    class Meta:
        model = TopicPollOption


class TopicPollVoteFactory(factory.django.DjangoModelFactory):
    poll_option = factory.SubFactory(TopicPollOptionFactory)
    voter = factory.SubFactory(UserFactory)

    class Meta:
        model = TopicPollVote



"""
    Forum defailt permissions
    =========================

    This module defines the default permissions that can be configured for a forum applications.

"""

from collections import OrderedDict

from django.utils.translation import gettext_lazy as _


class PermissionConfig:
    """ Defines the supported forum permissions.

    This class defines the default configuration of forum permissions. It is used to create the
    related ``ForumPermission`` instances on a syncdb or migrate operation. Moreover the permission
    scopes are used in admin pages to display permission forms in a readable way.

    """

    permissions = [
        # Forums
        {
            'codename': 'can_see_forum',
            'label': _('Can see forum'),
            'scope': 'forum',
        },
        {
            'codename': 'can_read_forum',
            'label': _('Can read forum'),
            'scope': 'forum',
        },

        # Topics & posts
        {
            'codename': 'can_start_new_topics',
            'label': _('Can start new topics'),
            'scope': 'conversation',
        },
        {
            'codename': 'can_reply_to_topics',
            'label': _('Can reply to topics'),
            'scope': 'conversation',
        },
        {
            'codename': 'can_post_announcements',
            'label': _('Can post announcements'),
            'scope': 'conversation',
        },
        {
            'codename': 'can_post_stickies',
            'label': _('Can post stickies'),
            'scope': 'conversation',
        },
        {
            'codename': 'can_delete_own_posts',
            'label': _('Can delete own posts'),
            'scope': 'conversation',
        },
        {
            'codename': 'can_edit_own_posts',
            'label': _('Can edit own posts'),
            'scope': 'conversation',
        },
        {
            'codename': 'can_post_without_approval',
            'label': _('Can post without approval'),
            'scope': 'conversation',
        },

        # Polls
        {
            'codename': 'can_create_polls',
            'label': _('Can create polls'),
            'scope': 'polls',
        },
        {
            'codename': 'can_vote_in_polls',
            'label': _('Can vote in polls'),
            'scope': 'polls',
        },

        # Attachments
        {
            'codename': 'can_attach_file',
            'label': _('Can attach file'),
            'scope': 'attachments',
        },
        {
            'codename': 'can_download_file',
            'label': _('Can download file'),
            'scope': 'attachments',
        },

        # Moderation
        {
            'codename': 'can_lock_topics',
            'label': _('Can lock topics'),
            'scope': 'moderation',
        },
        {
            'codename': 'can_move_topics',
            'label': _('Can move topics'),
            'scope': 'moderation',
        },
        {
            'codename': 'can_edit_posts',
            'label': _('Can edit posts'),
            'scope': 'moderation',
        },
        {
            'codename': 'can_delete_posts',
            'label': _('Can delete posts'),
            'scope': 'moderation',
        },
        {
            'codename': 'can_approve_posts',
            'label': _('Can approve posts'),
            'scope': 'moderation',
        },
        {
            'codename': 'can_reply_to_locked_topics',
            'label': _('Can add posts in locked topics'),
            'scope': 'moderation',
        },
    ]

    scopes = list(OrderedDict.fromkeys([p['scope'] for p in permissions]))

    def __getitem__(self, key):
        return self._permissions_dict[key]

    def get(self, key, default=None):
        return self._permissions_dict.get(key, default)

    @property
    def _permissions_dict(self):
        if not hasattr(self, '_permissions_per_codename'):
            self._permissions_per_codename = {p['codename']: p for p in self.permissions}
        return self._permissions_per_codename



# Generated by Django 3.1.2 on 2020-12-20 22:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('forum_conversation', '0012_auto_20200423_1049'),
    ]

    operations = [
        migrations.AlterField(
            model_name='topic',
            name='created',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation date'),
        ),
        migrations.AlterField(
            model_name='topic',
            name='last_post_on',
            field=models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='Last post added on'),
        ),
        migrations.AlterField(
            model_name='topic',
            name='updated',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='Update date'),
        ),
        migrations.AddIndex(
            model_name='topic',
            index=models.Index(fields=['type', 'last_post_on'], name='forum_conve_type_cc96d0_idx'),
        ),
    ]



"""
    Forum polls validators
    ======================

    This module defines validators provided by the ``forum_polls`` application.

"""

from django.core import validators

from machina.conf import settings as machina_settings


poll_max_options = [
    validators.MinValueValidator(1),
    validators.MaxValueValidator(machina_settings.POLL_MAX_OPTIONS_PER_USER),
]



# Generated by Django 2.2.10 on 2020-04-23 15:49

from django.db import migrations
import machina.core.validators
import machina.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('forum_conversation', '0011_remove_post_poster_ip'),
    ]

    operations = [
        migrations.AlterField(
            model_name='post',
            name='content',
            field=machina.models.fields.MarkupTextField(no_rendered_field=True, validators=[machina.core.validators.MarkupMaxLengthValidator(None)], verbose_name='Content'),
        ),
    ]



"""
    Forum moderation forms
    ======================

    This module defines forms provided by the ``forum_moderation`` application.

"""

from django import forms
from django.utils.translation import gettext_lazy as _

from machina.core.db.models import get_model
from machina.core.loading import get_class
from machina.forms.widgets import SelectWithDisabled


Forum = get_model('forum', 'Forum')

PermissionHandler = get_class('forum_permission.handler', 'PermissionHandler')


class TopicMoveForm(forms.Form):
    """ Allows to move a topic. """

    forum = forms.ChoiceField(label=_('Select a destination forum'), widget=SelectWithDisabled)
    lock_topic = forms.BooleanField(label=_('Lock topic'), required=False)

    def __init__(self, *args, **kwargs):
        self.topic = kwargs.pop('topic', None)
        self.user = kwargs.pop('user', None)
        self.perm_handler = PermissionHandler()

        super().__init__(*args, **kwargs)

        self.allowed_forums = self.perm_handler.get_target_forums_for_moved_topics(self.user)
        forum_choices = []

        for f in self.allowed_forums:
            if f.is_category or f.id == self.topic.forum.id:
                forum_choices.append((
                    f.id,
                    {
                        'label': '{} {}'.format('-' * f.margin_level, f.name),
                        'disabled': True,
                    }
                ))
            else:
                forum_choices.append((f.id, '{} {}'.format('-' * f.margin_level, f.name)))

        if self.topic.is_locked:
            self.fields['lock_topic'].initial = True

        self.fields['forum'].choices = forum_choices

    def clean_forum(self):
        forum_id = self.cleaned_data['forum']

        if forum_id:
            forum = Forum.objects.get(pk=forum_id)
            if forum.is_category or forum.is_link or forum.id == self.topic.forum.id:
                raise forms.ValidationError('You cannot select this forum as a destination')

        return forum



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-05-24 01:36
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('forum', '0005_auto_20170504_2113'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='forum',
            name='posts_count',
        ),
        migrations.RemoveField(
            model_name='forum',
            name='topics_count',
        ),
    ]



from importlib import import_module

from django.apps import apps
from django.apps.config import MODELS_MODULE_NAME
from django.core.exceptions import AppRegistryNotReady


# The following is mainly inspired from the model loading tools provided
# by django-oscar to handle overridable model classes.


def get_model(app_label, model_name):
    """ Given an app label and a model name, returns the corresponding model class. """
    try:
        return apps.get_model(app_label, model_name)
    except AppRegistryNotReady:
        if apps.apps_ready and not apps.models_ready:
            # If the models module of the considered app has not been loaded yet,
            # we try to import it manually in order to retrieve the model class.
            # This trick is usefull in order to get the model class during the
            # ``apps.populate()`` call and the execution of related methods of AppConfig
            # instances (eg. ``ready``).

            # Firts, the config of the consideration must be retrieved
            app_config = apps.get_app_config(app_label)
            # Then the models module is manually imported
            import_module('%s.%s' % (app_config.name, MODELS_MODULE_NAME))
            # Finally, tries to return the registered model class
            return apps.get_registered_model(app_label, model_name)
        else:
            raise


def is_model_registered(app_label, model_name):
    """ Checks whether the given model is registered or not.

    It is usefull to prevent Machina models for being registered if they are overridden by local
    apps.

    """
    try:
        apps.get_registered_model(app_label, model_name)
    except LookupError:
        return False
    else:
        return True


def model_factory(abstract_class):
    """ Given an abstract class, constructs the model that inherits from this class only if a model
        with the same (app label, model name) was not already in the app registry.
    """
    app_label = abstract_class.Meta.app_label
    model_name = abstract_class.__name__.replace('Abstract', '')

    if not is_model_registered(app_label, model_name):
        return type(str(model_name), (abstract_class, ), {'__module__': __name__, })



"""
    Forum polls abstract models
    ===========================

    This module defines abstract models provided by the ``forum_polls`` application.

"""

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from machina.apps.forum_conversation.forum_polls import validators
from machina.models.abstract_models import DatedModel


class AbstractTopicPoll(DatedModel):
    """ Represents a poll embedded in a forum topic. """

    topic = models.OneToOneField(
        'forum_conversation.Topic', related_name='poll', on_delete=models.CASCADE,
        verbose_name=_('Topic'),
    )

    # A poll is defined by a single question
    question = models.CharField(max_length=255, verbose_name=_('Poll question'))

    # A poll can have a duration
    duration = models.PositiveIntegerField(
        verbose_name=_('Poll duration, in days'), blank=True, null=True,
    )

    # Users can possibly select more than one option associated with a poll
    max_options = models.PositiveSmallIntegerField(
        verbose_name=_('Maximum number of poll options per user'),
        validators=validators.poll_max_options, default=1,
    )

    # Are users allowed to change their votes ?
    user_changes = models.BooleanField(verbose_name=_('Allow vote changes'), default=False)

    # Hide results to everyone except topic creator ?
    hide_results = models.BooleanField(verbose_name=_('Hide results'), default=False)

    class Meta:
        abstract = True
        app_label = 'forum_polls'
        ordering = ['-updated', ]
        get_latest_by = 'updated'
        verbose_name = _('Topic poll')
        verbose_name_plural = _('Topic polls')

    def __str__(self):
        return '{}'.format(self.topic.subject)

    @cached_property
    def votes(self):
        """ Returns all the votes related to this topic poll. """
        votes = []
        for option in self.options.all():
            votes += option.votes.all()
        return votes


class AbstractTopicPollOption(models.Model):
    """ Represents a poll option. """

    poll = models.ForeignKey(
        'forum_polls.TopicPoll', related_name='options', on_delete=models.CASCADE,
        verbose_name=_('Poll'),
    )
    text = models.CharField(max_length=255, verbose_name=_('Poll option text'))

    class Meta:
        abstract = True
        app_label = 'forum_polls'
        verbose_name = _('Topic poll option')
        verbose_name_plural = _('Topic poll options')

    def __str__(self):
        return '{} - {}'.format(self.poll, self.text)

    @cached_property
    def percentage(self):
        """ Returns the percentages of votes associated with the considered option. """
        return (self.votes.count() / (len(self.poll.votes) or 1)) * 100


class AbstractTopicPollVote(models.Model):
    """ Represents a poll vote. """

    voter = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name='poll_votes', blank=True, null=True,
        on_delete=models.CASCADE, verbose_name=_('Voter'),
    )
    anonymous_key = models.CharField(
        max_length=100, verbose_name=_('Anonymous user forum key'), blank=True, null=True,
    )

    poll_option = models.ForeignKey(
        'forum_polls.TopicPollOption', related_name='votes', on_delete=models.CASCADE,
        verbose_name=_('Poll option'),
    )
    timestamp = models.DateTimeField(auto_now_add=True, verbose_name=_('Vote\'s date'))

    class Meta:
        abstract = True
        app_label = 'forum_polls'
        verbose_name = _('Topic poll vote')
        verbose_name_plural = _('Topic poll votes')

    def __str__(self):
        return '{} - {}'.format(self.poll_option, self.voter)

    def clean(self):
        """ Validates the considered instance. """
        super().clean()

        # At least a poster (user) or a session key must be associated with
        # the vote instance.
        if self.voter is None and self.anonymous_key is None:
            raise ValidationError(_('A user id or an anonymous key must be used.'))
        if self.voter and self.anonymous_key:
            raise ValidationError(_('A user id or an anonymous key must be used, but not both.'))
